# This file is part of "austin-tui" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# austin-tui is top-like Web for Austin.
#
# Copyright (c) 2018-2020 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from enum import Enum
import json
import sys
from typing import Any, List, Optional, Type
import weakref

from aiohttp import web
from aiohttp.test_utils import unused_port
from austin import AustinError
from austin.aio import AsyncAustin
from austin.cli import AustinArgumentParser, AustinCommandLineError
from austin_web.data import DataPool, WebFrame
from austin_web.html import load_compile, load_site
from halo import Halo
from pyfiglet import Figlet


if sys.platform == "win32":
    asyncio.set_event_loop(asyncio.ProactorEventLoop())


class AustinWebError(AustinError):
    """AustinWeb own generic error."""

    pass


class AustinWebArgumentParser(AustinArgumentParser):
    """AustinWeb command line parser."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(name="austin-tui", full=False, alt_format=False)

        # ---- Serve command ----
        self.add_argument(
            "-S",
            "--serve",
            help="Serve Austin Web for live Austin data visualisation.",
            action="store_true",
        )

        self.add_argument(
            "-H",
            "--host",
            help="Set the host to serve on. Defaults to localhost.",
            type=str,
            default="localhost",
        )
        self.add_argument(
            "-P",
            "--port",
            help="Set the port to serve on. Defaults to an ephemaral port.",
            type=int,
            default=0,
        )

        # ---- Compile command ----
        self.add_argument(
            "-c",
            "--compile",
            help="Compile Austin data into an HTML flame graph visualisation "
            "into the given output file.",
            type=str,
        )


class AustinWebMode(Enum):
    """AustinWeb mode."""

    SERVE = 0
    COMPILE = 1


class AustinWeb(AsyncAustin):
    """AustinWeb subclass of AsyncAustin."""

    def __init__(self, args: Optional[List[str]] = None) -> None:
        super().__init__()

        self._args = AustinWebArgumentParser().parse_args(args)
        if self._args.compile and self._args.serve:
            raise AustinCommandLineError("Incompatible options: compile and serve.", -1)

        self._mode = (
            AustinWebMode.COMPILE if self._args.compile else AustinWebMode.SERVE
        )
        self._pools: weakref.WeakSet = weakref.WeakSet()
        self._pool: Optional[DataPool] = None
        self._runner: Optional[web.AppRunner] = None
        self._global_stats: Optional[str] = None
        self._spinner: Optional[Halo] = None

    def on_ready(self, *args: Any, **kwargs: Any) -> None:
        """Austin ready callback."""
        if self._mode == AustinWebMode.SERVE:
            asyncio.get_event_loop().create_task(self.start_server())
        else:
            self._spinner = Halo(text="Sampling", spinner="dots")
            self._spinner.start()
            self._pool = DataPool(self)

    def on_sample_received(self, text: str) -> None:
        """Austin sample received callback."""
        if self._mode == AustinWebMode.SERVE:
            if not self._pools:
                return
            frame = WebFrame.parse(text)
            for data_pool in self._pools:
                data_pool.add(frame)
        else:
            if not self._pool:
                raise AustinWebError("Data pool is unexpectedly missing")
            self._pool.add(WebFrame.parse(text))

    def on_terminate(self, stats: str) -> None:
        """Austin terminate callback."""
        self._global_stats = stats
        if self._mode == AustinWebMode.COMPILE:
            if self._spinner:
                self._spinner.stop()

            self.compile()

            asyncio.get_event_loop().stop()

    def new_data_pool(self) -> DataPool:
        """Make new data pool for incoming request."""
        data_pool = DataPool(self)
        self._pools.add(data_pool)
        return data_pool

    def discard_data_pool(self, data_pool: DataPool) -> None:
        """Discard no longer used data pool."""
        self._pools.discard(data_pool)

    def compile(self) -> None:
        """Compile collected samples."""
        if self._pool is None:
            return

        with open(self._args.compile, "w") as fout:
            fout.write(
                load_compile(
                    data=json.dumps(self._pool.data.to_dict()),
                    profile_type="Memory" if self._args.memory else "Time",
                )
            )
            print(f"✨🧁✨ Samples compiled into {self._args.compile}")

    async def handle_home(self, request: web.Request) -> web.Response:
        """Home page handler."""
        return web.Response(body=self.html, content_type="text/html")

    async def handle_websocket(self, request: web.Request) -> web.WebSocketResponse:
        """Web socket handler."""
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        data_pool = self.new_data_pool()

        try:
            payload = {
                "type": "info",
                "pid": self.get_process().pid,
                "command": self.get_command_line(),
                "metric": "m" if self._args.memory else "t",
            }

            await ws.send_str(json.dumps(payload))

            async for msg in ws:
                if not await data_pool.send(ws):
                    break

            self.discard_data_pool(data_pool)

        except AustinError:
            pass

        return ws

    async def start_server(self) -> None:
        """Start the web server asynchronously."""
        app = web.Application()
        app.add_routes(
            [web.get("/", self.handle_home), web.get("/ws", self.handle_websocket)]
        )

        port = self._args.port or unused_port()
        host = self._args.host

        self.html = load_site()

        self._runner = web.AppRunner(app)
        if not self._runner:
            raise AustinWebError("Cannot create web app runner.")
        await self._runner.setup()
        site = web.TCPSite(self._runner, host, port)
        await site.start()

        print(Figlet(font="speed", width=240).renderText("* Austin Web *"))
        print(
            f"⏲️ Sampling process with PID {self.get_process().pid} "
            f"({self.get_command_line()})"
        )
        print(f"🏃 Austin Web is running on http://{host}:{port}. Press Ctrl+C to stop.")

    async def stop_server(self) -> None:
        """Stop the web server asynchronously."""
        if self._runner:
            await self._runner.cleanup()

    async def start(self, args: List[str]) -> None:
        """Start austin and catch any exceptions."""
        try:
            await super().start(args)
        except AustinError as e:
            raise KeyboardInterrupt("Failed to start Austin") from e

    def run(self) -> None:
        """Run Austin Web."""
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            asyncio.set_event_loop(asyncio.new_event_loop())
            loop = asyncio.get_event_loop()

        try:
            loop.create_task(self.start(AustinWebArgumentParser.to_list(self._args)))
            loop.run_forever()
        except KeyboardInterrupt as e:
            if e.__cause__:
                print(
                    "❌ Austin failed to start. Please ensure that the Austin binary\n"
                    "is on the PATH environment variable and that the command line\n"
                    "arguments are correct."
                )
            pass
        finally:
            self.shutdown()

    def shutdown(self) -> None:
        """Shutdown Austin Web."""
        asyncio.get_event_loop().run_until_complete(self.stop_server())

        try:
            all_tasks = asyncio.all_tasks  # Python 3.7+
        except AttributeError:
            all_tasks = asyncio.Task.all_tasks  # Python 3.6

        try:
            for task in all_tasks():
                task.cancel()
        except RuntimeError:
            return

        pending = [task for task in all_tasks() if not task.done()]
        if pending:
            try:
                done, _ = asyncio.get_event_loop().run_until_complete(
                    asyncio.wait(pending)
                )
                for t in done:
                    res = t.result()
                    if res:
                        print(res)
            except (AustinError, asyncio.CancelledError):
                pass

        if self._global_stats:
            print(self._global_stats)


def _main(cls: Type[AustinWeb], args: List[str]) -> None:
    cls(args).run()


def main() -> None:
    """The main function."""
    try:
        _main(AustinWeb, sys.argv[1:])
    except AustinCommandLineError as e:
        message, *code = e.args
        print(message)
        exit(code[0] if code else -1)


if __name__ == "__main__":
    main()



import json
from typing import Any, Dict, List, Optional

from aiohttp import web
from austin.aio import AsyncAustin
from austin.stats import Frame, Sample
import psutil


class WebFrame:
    """Frame class designed to work nicely with d3-flame-graph."""

    __slots__ = ["name", "value", "children", "index", "parent", "height"]

    def __init__(self, name: Any, value: int) -> None:
        self.name: str = str(name)
        self.value: int = value
        self.children: List["WebFrame"] = []
        self.index: Dict[str, "WebFrame"] = {}
        self.parent: Optional["WebFrame"] = None
        self.height: int = 1

    def __add__(self, other: "WebFrame") -> "WebFrame":
        """Update the values by merging the right frame into the left one."""
        if self.name != other.name:
            if not self.parent:
                raise ValueError("Parent missing in non-root WebFrame")
            self.parent.add_child(other)
            self.parent.height = max(self.height, other.height) + 1
        else:
            self.value += other.value
            if other.height > self.height:
                self.height = other.height
            for child in other.children:
                try:
                    self.index[child.name] += child
                except KeyError:
                    self.add_child(child)

        return self

    def add_child(self, frame: "WebFrame") -> None:
        """Add child frame to the frame."""
        self.index[frame.name] = frame
        frame.parent = self
        self.children.append(frame)

    @staticmethod
    def parse(text: str) -> "WebFrame":
        """Create a frame from an Austin collapsed sample."""

        def build_frame(frames: List[Frame]) -> "WebFrame":
            first, *tail = frames
            frame = WebFrame(str(first), sample.metrics.time)
            if tail:
                child = build_frame(tail)
                frame.add_child(child)
                frame.height = child.height + 1
            else:
                frame.value = sample.metrics.time
            return frame

        sample = Sample.parse(text)
        process_frame = WebFrame(sample.pid, sample.metrics.time)
        thread_frame = WebFrame(sample.thread, sample.metrics.time)

        if sample.frames:
            thread_frame.add_child(build_frame(sample.frames))

        thread_frame.height = len(sample.frames) + 1

        process_frame.add_child(thread_frame)
        process_frame.height = thread_frame.height + 1

        root = WebFrame.new_root()
        root.add_child(process_frame)
        root.value = process_frame.value
        root.height = process_frame.height + 1

        return root

    @staticmethod
    def new_root() -> "WebFrame":
        """Create a new root frame."""
        return WebFrame("root", 0)

    def to_dict(self) -> dict:
        """Return the frame data in the form of a ``dict``."""
        return {
            "name": self.name,
            "value": self.value,
            "children": [c.to_dict() for c in self.children],
        }

    def __repr__(self) -> str:
        """The frame representation of its fields."""
        return (
            type(self).__name__
            + f"({ {s: getattr(self, s) for s in ['name', 'value', 'children', 'height']} })"
        )


class DataPool:
    """Data collection pool to serve each client handler.

    When the data is sent, the content of the pool is reset to just a root
    node. Each received sample should be added to each data pool so that each
    client can get the data that was seen from the moment they connected.
    """

    def __init__(self, austin: AsyncAustin) -> None:
        self._austin = austin
        self.max = 0
        self.data = WebFrame.new_root()
        self.samples = 0

    def add(self, frame: WebFrame) -> None:
        """Add frame to the data pool."""
        self.data += frame
        self.samples += 1

    async def send(self, ws: web.WebSocketResponse) -> bool:
        """Send profiling data to websocket asynchronously.

        Returns ``True`` on success, ``False`` otherwise.
        """
        try:
            cpu = self._austin.get_child_process().cpu_percent()
            memory = self._austin.get_child_process().memory_full_info()[0] >> 20
        except psutil.NoSuchProcess:
            return False

        data = self.data.to_dict()

        if self.data.height > self.max:
            self.max = self.data.height

        payload = {
            "type": "sample",
            "data": data,
            "height": self.max,
            "samples": self.samples,
            "cpu": cpu,
            "memory": memory,
        }

        await ws.send_str(json.dumps(payload))

        self.data = WebFrame.new_root()

        return True



"""HTML resource handling utilities."""

from typing import Optional, Tuple

from importlib_resources import files


def get_resource(name: str) -> str:
    """Load a resource file from the submodule ``austin_web.html``."""
    return files("austin_web.html").joinpath(name).read_text(encoding="utf-8")


def _find_marker(
    text: str, ldel: str, rdel: str, begin: int
) -> Optional[Tuple[int, int, str]]:
    begin = text.find(ldel, begin)
    if begin < 0:
        return None
    end = text.find(rdel, begin) + len(rdel)

    return begin, end, text[begin:end]


def replace_references(text: str) -> str:
    """Replace {{ reference }} with the content of the referenced resource."""
    begin = 0
    while True:
        marker = _find_marker(text, "{{", "}}", begin)
        if not marker:
            break

        begin, end, placeholder = marker

        reference = placeholder[2:-2].strip()
        resource = get_resource(reference)

        text = text.replace(placeholder, replace_references(resource))

        begin += len(resource)

    return text


def replace_links(text: str) -> str:
    """Replace [[ link ]] with a link to the referenced resource."""
    begin = 0
    while True:
        marker = _find_marker(text, "[[", "]]", begin)
        if not marker:
            break

        begin, end, placeholder = marker

        reference = placeholder[2:-2].strip()
        link = "file://" + get_resource(reference)

        text = text.replace(placeholder, link)

        begin += len(link)

    return text


def replace_placeholders(text: str, **kwargs: str) -> str:
    """Replace occurrences of ((% placeholder %)) with the given keyword arguments."""
    if not kwargs:
        return text

    begin = 0
    while True:
        marker = _find_marker(text, r"((%", r"%))", begin)
        if not marker:
            break

        begin, end, placeholder = marker

        key = placeholder[3:-3].strip()
        value = kwargs.get(key)

        if not value:
            begin += 3
            continue

        text = text.replace(placeholder, value)

        begin += len(value)

    return text


def _load_and_resolve_resource(name: str, **kwargs: str) -> str:
    return replace_placeholders(
        replace_links(replace_references(get_resource(name))), **kwargs
    )


def load_site() -> str:
    """Load the site index page."""
    return _load_and_resolve_resource("index.html")


def load_compile(data: str, profile_type: str) -> str:
    """Load the compiler page.

    The ``data`` string argument is a serialised JSON object. The
    ``profile_type`` should be either ``Time`` or ``Memory``.
    """
    return _load_and_resolve_resource(
        "austin_web.html",
        data=data,
        profile_type=profile_type,
        label=profile_type.lower(),
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['austin_web', 'austin_web.html']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.6.2,<4.0.0',
 'austin-python>=0.1.0,<0.2.0',
 'halo>=0.0.29,<0.0.30',
 'importlib_resources>=3.0.0,<4.0.0',
 'pyfiglet>=0.8.post1,<0.9']

entry_points = \
{'console_scripts': ['austin-web = austin_web.__main__:main']}

setup_kwargs = {
    'name': 'austin-web',
    'version': '0.1.0',
    'description': 'Flame graph web application for Austin',
    'long_description': '<p align="center">\n  <br>\n  <img width="320px" src="art/logo.png" alt="Austin Web">\n  <br>\n</p>\n\n<h3 align="center">A Modern Web Interface for Austin</h3>\n\n<p align="center">\n  <img src="https://upload.wikimedia.org/wikipedia/commons/3/3a/Tux_Mono.svg"\n       height="24px" />\n  &nbsp;&nbsp;&nbsp;&nbsp;\n  <img src="https://upload.wikimedia.org/wikipedia/commons/f/fa/Apple_logo_black.svg"\n       height="24px" />\n  &nbsp;&nbsp;&nbsp;&nbsp;\n  <img src="https://upload.wikimedia.org/wikipedia/commons/2/2b/Windows_logo_2012-Black.svg"\n       height="24px" />\n</p>\n\n<p align="center">\n  <a href="https://github.com/P403n1x87/austin-web/actions?workflow=Tests">\n    <img src="https://github.com/P403n1x87/austin-web/workflows/Tests/badge.svg"\n         alt="GitHub Actions: Tests">\n  </a>\n  <a href="https://travis-ci.com/P403n1x87/austin-web">\n    <img src="https://travis-ci.com/P403n1x87/austin-web.svg?token=fzW2yzQyjwys4tWf9anS"\n         alt="Travis CI">\n  </a>\n  <a href="https://codecov.io/gh/P403n1x87/austin-web">\n    <img src="https://codecov.io/gh/P403n1x87/austin-web/branch/master/graph/badge.svg"\n         alt="Codecov">\n  </a>\n  <a href="https://pypi.org/project/austin-web/">\n    <img src="https://img.shields.io/pypi/v/austin-web.svg"\n         alt="PyPI">\n  </a>\n  <a href="https://github.com/P403n1x87/austin-web/blob/master/LICENSE.md">\n    <img src="https://img.shields.io/badge/license-GPLv3-ff69b4.svg"\n         alt="LICENSE">\n  </a>\n</p>\n\n<p align="center">\n  <a href="#synopsis"><b>Synopsis</b></a>&nbsp;&bull;\n  <a href="#installation"><b>Installation</b></a>&nbsp;&bull;\n  <a href="#usage"><b>Usage</b></a>&nbsp;&bull;\n  <a href="#compatibility"><b>Compatibility</b></a>&nbsp;&bull;\n  <a href="#contribute"><b>Contribute</b></a>\n</p>\n\n<p align="center">\n  <a href="https://www.buymeacoffee.com/Q9C1Hnm28"\n     target="_blank">\n  <img src="https://www.buymeacoffee.com/assets/img/custom_images/orange_img.png"\n       alt="Buy Me A Coffee" />\n  </a>\n</p>\n\n# Synopsis\n\nAustin Web is a modern web interface for Austin, based on\n[D3.js](https://d3js.org/) and [tailwindcss](https://tailwindcss.com/). It is\nyet another example of how to use Austin to make a visual profiling tool for\nPython. The flame graph is generated using\n[d3-flame-graph](https://github.com/spiermar/d3-flame-graph).\n\n<p align="center">\n  <img src="art/austin-web-serve.gif"\n       style="box-shadow: #111 0px 0px 16px;" />\n</p>\n\nAustin Web offers two main functionalities. The default one is to serve a web\npage that allows you to have a live view of the metrics collected by Austin. The\nvisualisation is a _live_ flame graph in your browser that refreshes every 3\nseconds with newly collected data. Hence, Austin Web can also be used for\n_remote_ profiling.\n\nYou can also run Austin Web in _compile_ mode to generate a static flame graph\nHTML page, much like\n[flamegraph.pl](https://github.com/brendangregg/FlameGraph), but with the full\nAustin Web UI around it.\n\n\n# Installation\n\nAustin Web can be installed from PyPI simply with\n\n~~~ bash\npip install austin-web\n~~~\n\n> **NOTE** Austin Web relies on the\n> [Austin](https://github.com/P403n1x87/austin) binary being available from the\n> `PATH` environment variable. So make sure that Austin is properly installed on\n> your system. See\n> [Austin installation](https://github.com/P403n1x87/austin#installation)\n> instruction for more details on how to get Austin installed on your platform.\n\n\n# Usage\n\nYou can run Austin Web simply with\n\n~~~ bash\naustin-web python3 myscript.py\n~~~\n\nto start serving on localhost over an ephemeral port. If `myscript.py` is an\nexecutable script, you can simply do\n\n~~~ bash\naustin-web myscript.py\n~~~\n\nIf you want to specify the host and the port, you can pass the `--host` and\n`--port` options to the command line. For example, to serve for the World on\nport 5050, use\n\n~~~ bash\naustin-web --host 0.0.0.0 --port 5050 python3 myscript.py\n~~~\n\nIf you want to compile the collected metrics into a static HTML page, you can\nrun Austin Web in compile mode by passing the `--compile` option, followed by\nthe destination file name, e.g.\n\n~~~ bash\naustin-web --compile output.html python3 myscript.py\n~~~\n\nLike Austin, you can use Austin Web to profile any running Python application.\nFor example, to profile a WSGI server and all its child processes, get hold of\nits PID and do\n\n~~~ bash\nsudo austin-web -Cp <pid>\n~~~\n\n\n# Compatibility\n\nAustin Web has been tested with Python 3.6-3.9 and is known to work on\n**Linux**, **MacOS** and **Windows**.\n\nAustin Web is known to have some minor issues on Windows. When started in serve\nmode, pressing `Ctrl+C` might not actually stop Austin Web.\n\nSince Austin Web uses Austin to collect samples, the same note applies here:\n\n> Due to the **System Integrity Protection** introduced in **MacOS** with El\n> Capitan, Austin cannot profile Python processes that use an executable located\n> in the `/bin` folder, even with `sudo`. Hence, either run the interpreter from\n> a virtual environment or use a Python interpreter that is installed in, e.g.,\n> `/Applications` or via `brew` with the default prefix (`/usr/local`). Even in\n> these cases, though, the use of `sudo` is required.\n\n\n# Contribute\n\nIf you want to help with the development, then have a look at the open issues\nand have a look at the [contributing guidelines](CONTRIBUTING.md) before you\nopen a pull request.\n\nYou can also contribute to the development of the Austin Web by becoming a\nsponsor and/or by [buying me a coffee](https://www.buymeacoffee.com/Q9C1Hnm28)\non BMC or by chipping in a few pennies on\n[PayPal.Me](https://www.paypal.me/gtornetta/1).\n\n<p align="center">\n  <a href="https://www.buymeacoffee.com/Q9C1Hnm28"\n     target="_blank">\n  <img src="https://www.buymeacoffee.com/assets/img/custom_images/orange_img.png"\n       alt="Buy Me A Coffee" />\n  </a>\n</p>\n',
    'author': 'Gabriele N. Tornetta',
    'author_email': 'phoenix1987@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/P403n1x87/austin-web',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
